use std::error::Error;
use flo_curves::*;
use flo_curves::{bezier, Coord2};

use super::utility::print_info;

/*
stinkhead7ds easing curve description:

e.g. going from 120 to 140
easing function is cubic bezier going from 0.0 to 1.0

140-120=20 increments
0.0 is +0 and 1.0 is +20

console has no floating point and fixed point will overflow 16-bit
each curve tailored for use case. for camera motion:
- generate the cubic bezier easing curve in my custom branch of reskit, converting 0.0-1.0 floating point spans to x/128 fractions equivalent to x/100 fractions.
  literal grade school shit i realised i wasted a weekend re-deriving from first principles. these fparts even export as single bytes, so that's 3 less bytes per point
  on the curve than if i were sticking with 16.16 fix. for a 1 second animation, 60 points are taken on the 0.0-1.0 curve, for a 2 second animation, 120 points are taken, etc.
  equivalent PAL framerate curves can be export in the same curve file.
- the camera animates moves by setting a new destination point and performing interpolation of intermediate positions along the curve - just like css easing functions, or
  unity/unreal bezier curves. you restrict camera new position to +/- powers of two in any axis to eliminate mul[s/u].w. e.g. camera moves +2, +4, +8.... in the x direction.
- for each axis, new position - current position = span of movement. so if you're starting at x = 30 and moving to x = 46, you're going +16 in the x direction
- while the camera is animating, interpolate positions by reading the fractional on the curve for the current frame and doing fractional * span / 128. since new camera positions
  in any axis are kept power of two this can all be done using asl/asr. take the remainder with the and 2^127 trick and bump the resulting value up 1 if the modulo is >= 64
- ✨ cubic bezier animated camera motion
*/


/**
 * Output a binary for a cubic bezier defined by the given four points and the
 * time interval defined as number of 1/60 jiffies.
 */
pub fn get_cubic_bezier( p0: (f64, f64), c0: (f64, f64), c1: (f64, f64), p1: (f64, f64), jiffies: u16 ) -> Result<Vec<u8>, Box<dyn Error>> {
    let mut result: Vec<u8> = Vec::new();

    // 2 bytes: Number of frames this curve spans
    result.extend( jiffies.to_be_bytes() );

    let curve = bezier::Curve::from_points(
        Coord2( p0.0, p0.1 ),
        ( Coord2( c0.0, c0.1 ), Coord2( c1.0, c1.1 ) ),
        Coord2( p1.0, p1.1 )
    );

    let mut current_jiffy = 0.0;
    // For each point in curve: 1 byte fractional point in the curve at the given frame
    for i in 0..jiffies {
        let point = curve.point_at_pos( current_jiffy );
        let x = point.1;

        // Round and convert x to n/100 fraction, then equivalent n/128 fraction
        let x = ( ( x * 100.0 ).round() / 100.0 ) * 100.0; // only get last two decimal places, then convert to whole number
        let x = x * 128.0 / 100.0; // x / 100 = y / 128
        let x = x.round() as u8;

        print_info( &format!( "at iteration {}, curve is {:?}, fractional format will be {}", i, point, x ) );
        current_jiffy += 1.0 / jiffies as f64;

        result.push( x );
    }

    Ok( result )
}