use std::{error::Error, collections::HashMap};
use image::GenericImageView;
use crate::reskit::{tileset::image_to_tiles, utility::symbol_to_pascal, cli::settings::TileOrder};
use super::converter::{TiledTilemap, Layer, SystemPlane, TiledTileset};

/**
 * Output the .bin and .pal file (using `tileset` tool to build it) containing each of the tiles
 * and palettes in the Tiled Editor tileset.
 */
pub fn get_tiles( tilemap: &TiledTilemap ) -> Result<(Vec<u8>, Vec<u8>), Box<dyn Error>> {
    let mut system_pals: [[u16; 16]; 4] = [
        [ 0; 16 ],
        [ 0; 16 ],
        [ 0; 16 ],
        [ 0; 16 ]
    ];
    let mut all_tiles: Vec<u8> = vec![0; 32];   // --system md, start with a blank buffer tile

    for tileset in &tilemap.tileset {
        let tiles_height = tileset.image.height() / 8;  // --system md
        let tiles_width = tileset.image.width() / 8;    // --system md

        // all this copy pasted code, it's so over
        if matches!( tileset.tile_order, TileOrder::Sprite ) {
            // Sprite iteration order
            for tile_x in 0..tiles_width {
                for tile_y in 0..tiles_height {
                    let tile_index = ( tile_y * tiles_width ) + tile_x;
                    let tile = tileset.image.clone().crop( tile_x * 8, tile_y * 8, 8, 8 );

                    // Fake palette (see below)
                    let mut fake: [u16; 16] = [ 0; 16 ];
                    let selected_pal = tileset.palettes[ tile_index as usize ];

                    let tile_bin = image_to_tiles(
                        &tile,
                        {
                            // Determine if palette is used here or it is a dummy palette
                            if let Some( selected_pal ) = selected_pal {
                                &mut system_pals[ selected_pal as usize ]
                            } else {
                                // Fake-a-palette
                                // You will get an error in get_tilemap if you try to use this palette-less tile
                                &mut fake
                            }
                        },
                        "tile"  // this is not set to "sprite" because we're only doing one tile at a time
                    );

                    if let Err( _ ) = tile_bin {
                        return Err( format!( "palette {:?} full (try moving tile {} to another palette)", selected_pal, tile_index ) )?
                    }

                    if let Ok( tile_bin ) = tile_bin {
                        all_tiles.extend( tile_bin );
                    }
                }
            }
        } else {
            // Tile iteration order
            for tile_y in 0..tiles_height {
                for tile_x in 0..tiles_width {
                    let tile_index = ( tile_y * tiles_width ) + tile_x;
                    let tile = tileset.image.clone().crop( tile_x * 8, tile_y * 8, 8, 8 );

                    // Fake palette (see below)
                    let mut fake: [u16; 16] = [ 0; 16 ];
                    let selected_pal = tileset.palettes[ tile_index as usize ];

                    let tile_bin = image_to_tiles(
                        &tile,
                        {
                            // Determine if palette is used here or it is a dummy palette
                            if let Some( selected_pal ) = selected_pal {
                                &mut system_pals[ selected_pal as usize ]
                            } else {
                                // Fake-a-palette
                                // You will get an error in get_tilemap if you try to use this palette-less tile
                                &mut fake
                            }
                        },
                        "tile"
                    );

                    if let Err( _ ) = tile_bin {
                        return Err( format!( "palette {:?} full (try moving tile {} to another palette)", selected_pal, tile_index ) )?
                    }

                    if let Ok( tile_bin ) = tile_bin {
                        all_tiles.extend( tile_bin );
                    }
                }
            }
        }
    }

    // Define result and write palettes
    let mut palettes: Vec<u8> = Vec::new();
    for pal in 0..system_pals.len() {
        for i in 0..system_pals[ pal ].len() {
            let bytes = system_pals[ pal ][ i ].to_be_bytes();
            for i in 0..2 {
                palettes.push( bytes[ i ] );
            }
        }
    }

    Ok( ( all_tiles, palettes ) )
}

/**
 * Get the .map file containing metatile definitions, as well as B and A metatile-based maps.
 *
 * Format:
 * <<< HEADER >>>
 * 2 bytes: The offset to the instances table
 * (n) 2 byte offsets: For the number of defined metatiles, 16-bit offsets into the file
 * pointing to the metatile definition. Index into this table = the metatile ID.
 *
 * <<< DEFINITIONS >>>
 * For each metatile definition:
 * 2 bytes: Width of the metatile, in 16-bit words.
 * 2 bytes: Height of the metatile, in 16-bit words.
 * (n) bytes: The nametable definition for the metatile. Stamp this into vram on each layer.
 *
 * <<< INSTANCES >>>
 * For B, then A layers:
 * 2 bytes: Number of metatile instances in this layer.
 * For each metatile instance:
 * 2 bytes: Metatile ID.
 * 2 bytes: X coordinate (world, divide by 8 to get tile coordinate).
 * 2 bytes: Y coordinate (world, divide by 8 to get tile coordinate).
 */
pub fn get_metatile_maps( tilemap: &TiledTilemap ) -> Result<Vec<u8>, Box<dyn Error>> {
    // Assemble header of offsets (as we go)
    let header_offset = ( tilemap.metatiles.len() * 2 ) + 2;
    let mut header: Vec<u8> = vec![];

    // Assemble definitions
    let mut definitions: Vec<u8> = vec![];
    for metatile in &tilemap.metatiles {
        let mut subdefinitions: Vec<u8> = vec![];
        subdefinitions.extend( metatile.width.to_be_bytes() );
        subdefinitions.extend( metatile.height.to_be_bytes() );

        let ( tilemap_index, palettes ) = metatile_tilemap_index( &metatile.source, &tilemap.tileset )?;
        for tile in &metatile.tiles {
            let palette = {
                if *tile == 0 {
                    0
                } else {
                    let palette = palettes.get( *tile as usize ).ok_or( "internal error: no correlation between tile index and palette index" )?;
                    if let Ok( palette ) = palette.ok_or( "" ) {
                        palette
                    } else {
                        println!( "{:?}", palettes );
                        return Err( format!( "invalid file: tile \"{}\" in metatile tilemap does not name reskit-palette attribute", tile ) )?;
                    }
                }
            };

            let nametable_entry: u16 = ( ( palette as u16 ) << 13 ) | ( ( tile + tilemap_index as u32 ) as u16 );
            subdefinitions.extend( nametable_entry.to_be_bytes() );
        }

        let this_offset = header_offset + definitions.len();
        header.extend( ( this_offset as u16 ).to_be_bytes() );

        definitions.extend( subdefinitions );
    }

    // Assemble instances
    let mut instances: Vec<u8> = vec![];

    let layer_b: Option<&Layer> = tilemap.layers.iter().find( | layer | matches!( layer, Layer::Metatile { system_plane: SystemPlane::MdPlaneB, tiles: _ } ) );
    if let Some( layer_b ) = layer_b {
        let tiles = match layer_b { Layer::Metatile { system_plane: _, tiles } => tiles };
        let mut subdefinitions: Vec<u8> = vec![];

        for tile_y in 0..tilemap.height {
            for tile_x in 0..tilemap.width {
                let item_at = tiles.get( ( tile_y * tilemap.width ) + tile_x ).ok_or( "internal error: tilemap does not correlate to width/height" )?;
                if *item_at != 0 {
                    // What index-ID was that?
                    let index_id = tilemap.metatiles.iter().position( | metatile | metatile.id == *item_at ).ok_or( "invalid file: metatile id not found" )?;

                    // Write the data to subdefinitions
                    subdefinitions.extend( ( index_id as u16 ).to_be_bytes() );
                    subdefinitions.extend( ( ( tile_x * 8 ) as u16 ).to_be_bytes() );
                    subdefinitions.extend( ( ( tile_y * 8 ) as u16 ).to_be_bytes() );
                }
            }
        }

        instances.extend( ( ( subdefinitions.len() / 6 ) as u16 ).to_be_bytes() );
        instances.extend( subdefinitions );
    } else {
        instances.extend( ( 0 as u16 ).to_be_bytes() );
    }

    let layer_a: Option<&Layer> = tilemap.layers.iter().find( | layer | matches!( layer, Layer::Metatile { system_plane: SystemPlane::MdPlaneA, tiles: _ } ) );
    if let Some( layer_a ) = layer_a {
        let tiles = match layer_a { Layer::Metatile { system_plane: _, tiles } => tiles };
        let mut subdefinitions: Vec<u8> = vec![];

        for tile_y in 0..tilemap.height {
            for tile_x in 0..tilemap.width {
                let item_at = tiles.get( ( tile_y * tilemap.width ) + tile_x ).ok_or( "internal error: tilemap does not correlate to width/height" )?;
                if *item_at != 0 {
                    // What index-ID was that?
                    let index_id = tilemap.metatiles.iter().position( | metatile | metatile.id == *item_at ).ok_or( "invalid file: metatile id not found" )?;

                    // Write the data to subdefinitions
                    subdefinitions.extend( ( index_id as u16 ).to_be_bytes() );
                    subdefinitions.extend( ( ( tile_x * 8 ) as u16 ).to_be_bytes() );
                    subdefinitions.extend( ( ( tile_y * 8 ) as u16 ).to_be_bytes() );
                }
            }
        }

        instances.extend( ( ( subdefinitions.len() / 6 ) as u16 ).to_be_bytes() );
        instances.extend( subdefinitions );
    } else {
        instances.extend( ( 0 as u16 ).to_be_bytes() );
    }

    // Write the final result
    let mut result: Vec<u8> = vec![];
    let offset_to_instances = header_offset + definitions.len();
    result.extend( ( offset_to_instances as u16 ).to_be_bytes() );
    result.extend( header );
    result.extend( definitions );
    result.extend( instances );

    Ok( result )
}

fn metatile_tilemap_index( target_source: &str, tilesets: &Vec<TiledTileset> ) -> Result<(usize, Vec<Option<u8>>), Box<dyn Error>> {
    // Everything starts at 1 for the blank buffer tile
    let mut index: usize = 1;

    for tileset in tilesets {
        if tileset.source == target_source {
            return Ok( ( index, tileset.palettes.clone() ) )
        } else {
            index += ( ( tileset.image.width() / 8 ) * ( tileset.image.height() / 8 ) ) as usize;
        }
    }

    Err( format!( "internal error: could not find metatile with target source \"{}\"", target_source ) )?
}

/**
 * Get the .lvc collision map (u8 sized for the map dimensions, collision areas are either 0 or 1)
 */
pub fn get_collision_map( tilemap: &TiledTilemap ) -> Result<Vec<u8>, Box<dyn Error>> {
    let mut result: Vec<u8> = Vec::new();

    // 2 bytes: Number of bounding boxes
    result.extend( ( tilemap.collision.len() as u16 ).to_be_bytes() );

    // For (Number of bounding boxes):
    for bounding_box in &tilemap.collision {
        // 2 bytes: X dimension
        result.extend( bounding_box.origin.x.to_be_bytes() );

        // 2 bytes: Y dimension
        result.extend( bounding_box.origin.y.to_be_bytes() );

        // 2 bytes: X2 dimension (x + width)
        result.extend( ( bounding_box.origin.x + bounding_box.size.width ).to_be_bytes() );

        // 2 bytes: Y2 dimension (y + height)
        result.extend( ( bounding_box.origin.y + bounding_box.size.height ).to_be_bytes() );
    }

    Ok( result )
}

pub fn get_objs( tilemap: &TiledTilemap, symbol_ids: &HashMap<String, u16> ) -> Result<Vec<u8>, Box<dyn Error>> {
    let mut result: Vec<u8> = Vec::new();

    // 2 bytes: Number of objects
    result.extend( ( tilemap.objects.len() as u16 ).to_be_bytes() );

    if tilemap.objects.len() > 0 {
        let archetype = tilemap.objects.first().ok_or( "internal error: no first object" )?;

        // 2 bytes: Object struct size
        result.extend( ( ( archetype.attributes.len() + 1 ) as u16 ).to_be_bytes() );

        // For (number of objects):
        for object in &tilemap.objects {
            // (Object struct size * 2): Data for object, ordered by --fields option, each one 16-bit field

            // First output the object ID
            let object_id = symbol_ids.get( &object.id ).ok_or( format!( "invalid file: undefined symbol \"{}\"", object.id ) )?;
            result.extend( object_id.to_be_bytes() );

            // Then output the attributes in the order provided via LinkedHashMap
            for ( _attribute, value ) in &object.attributes {
                let value: u16 = match value.as_str() {
                    "true" => 0x01,
                    "false" => 0x00,
                    all_else => {
                        if let Ok( valid_u16 ) = all_else.parse::<u16>() {
                            valid_u16
                        } else {
                            *symbol_ids.get( all_else ).ok_or( format!( "invalid file: undefined symbol \"{}\"", all_else ) )?
                        }
                    }
                };

                result.extend( value.to_be_bytes() );
            }
        }
    }

    Ok( result )
}

/**
 * Get the sprite table.
 */
pub fn get_sprites( tilemap: &TiledTilemap, sprite_ids: &HashMap<String, u16> ) -> Result<Vec<u8>, Box<dyn Error>> {
    let mut result: Vec<u8> = Vec::new();

    let tiles_before = {
        let mut total_tiles = 1;

        for tileset in &tilemap.tileset {
            // do not include the sprite metadata
            if tileset.sprite_metadata.is_none() {
                total_tiles += ( tileset.image.width() / 8 ) * ( tileset.image.height() / 8 );
            }
        }

        total_tiles
    };
    let mut tiles_after: u32 = 0;

    let sprites: Vec<&TiledTileset> = tilemap.tileset.iter().filter( | tileset | matches!( tileset.tile_order, TileOrder::Sprite ) ).collect();
    result.extend( ( sprites.len() as u16 ).to_be_bytes() );
    for sprite in sprites {
        let sprite_metadata = sprite.sprite_metadata.as_ref().ok_or( "internal error: tile order is sprite but no sprite data" )?;

        let sprite_id = sprite_ids.get( &sprite_metadata.id ).ok_or( format!( "invalid file: undefined sprite id \"{}\"", sprite_metadata.id ) )?;
        result.extend( sprite_id.to_be_bytes() );

        let tiles_across = sprite.image.width() / 8;
        let frames: u16 = ( tiles_across / ( sprite_metadata.width as u32 ) ) as u16;

        let tile_index_location = tiles_before + tiles_after;
        result.extend( ( tile_index_location as u16 ).to_be_bytes() );
        tiles_after += ( ( sprite_metadata.width as u16 * sprite_metadata.height as u16 ) * frames ) as u32;

        result.push( sprite_metadata.width );

        result.push( sprite_metadata.height );

        result.extend( sprite_metadata.palette.to_be_bytes() );

        // One animation per .tsx file
        // Animations run like a filmstrip across, never down
        result.extend( frames.to_be_bytes() );

        if let Some( jiffies ) = sprite_metadata.anim_interval {
            result.extend( jiffies.to_be_bytes() );
        } else {
            result.extend( ( 0 as u16 ).to_be_bytes() );
        };
    }

    Ok( result )
}

/**
 * Get a helper .asm or .c file that ties all the level components together
 */
pub fn get_code( tilemap: &TiledTilemap, level_name: &str, path_prefix: &str ) -> Result<String, Box<dyn Error>> {
    let version = env!( "CARGO_PKG_VERSION" );
    let level_label = symbol_to_pascal( level_name );
    let ( width, height ) = ( tilemap.width, tilemap.height );
    let num_tiles = {
        let mut total_tiles = 1;

        for tileset in &tilemap.tileset {
            total_tiles += ( tileset.image.width() / 8 ) * ( tileset.image.height() / 8 );     // --system md
        }

        total_tiles
    };
    let file = format!( r#"; Level definition file
; Generated by reskit v{version}

{level_label}Tiles:
    incbin '{path_prefix}{level_name}/tiles.bin'

{level_label}Palettes:
    incbin '{path_prefix}{level_name}/palettes.pal'

{level_label}Nametables:
    incbin '{path_prefix}{level_name}/nametables.map'

{level_label}Collision:
    incbin '{path_prefix}{level_name}/collision.lvc'

{level_label}Objects:
    incbin '{path_prefix}{level_name}/objects.obs'

{level_label}Sprites:
    incbin  '{path_prefix}{level_name}/sprites.spt'

{level_label}:
    dc.w    {width}, {height}, {num_tiles}
    dc.l    {level_label}Tiles
    dc.l    {level_label}Palettes
    dc.l    {level_label}Nametables
    dc.l    {level_label}Collision
    dc.l    {level_label}Objects
    dc.l    {level_label}Sprites"# );

    Ok( file )
}